#include "../../std.h"

#include <vector>

#include "../../misc/inttostring.h"
#include "../../misc/stringtoint.h"
#include "../../misc/log_time.h"

#include "srm_info.h"

bool SRMFileInfo::operator ==(SRM_URL srm_url) {
  const std::string proto_val =
       ((std::map<std::string,std::string>&)(srm_url.Options()))["protocol"];
  if (host == srm_url.Host() &&
      (!srm_url.PortDefined() || port == srm_url.Port()) &&
      ((proto_val == "") || ((protocol == "gssapi") == srm_url.GSSAPI())) &&
      version == srm_url.SRMVersion())
    return true;
  return false;
}

std::string SRMFileInfo::versionString() const {
  switch (version) {
    case SRM_URL::SRM_URL_VERSION_1: {
      return "1";
    }; break;
    case SRM_URL::SRM_URL_VERSION_2_2: {
      return "2.2";
    }; break;
    default: {
      return "";
    }
  }
  return "";
}

LockSimple SRMInfo::filelock = LockSimple();

SRMInfo::SRMInfo(std::string dir) {
  srm_info_filename = dir + "/srms.conf";
}

bool SRMInfo::getSRMFileInfo(SRMFileInfo& srm_file_info) {
  
  struct stat fileStat;
  int err = stat( srm_info_filename.c_str(), &fileStat ); 
  if (0 != err || fileStat.st_size == 0) {
    if (0 != err && errno != ENOENT)
      odlog(WARNING)<<"Error reading srm info file "<<srm_info_filename<<": "<<strerror(errno)<<std::endl;
    return false;
  }

  filelock.block();
  FILE * pFile;
  char mystring [fileStat.st_size+1];
  pFile = fopen (srm_info_filename.c_str(), "r");
  if (pFile == NULL) {
    odlog(WARNING)<<"Error opening srm info file "<<srm_info_filename<<": "<<strerror(errno)<<std::endl;
    filelock.unblock();
    return false;
  }

  // read in file
  char * res = fgets (mystring, sizeof(mystring), pFile);
  while (res) {
    std::string line(mystring);
    if (line.find('\n') != std::string::npos) line.resize(line.find('\n'));
    if (line.length() > 0 && line[0] == '#') {
      res = fgets (mystring, sizeof(mystring), pFile);
      continue;
    }
    std::string::size_type space_pos = line.find(' ');
    if (space_pos == 0 || space_pos == std::string::npos) {
      if (line.length() > 0) 
        odlog(WARNING)<<"Bad format detected in file "<<srm_info_filename<<", in line "<<line<<std::endl;
      res = fgets (mystring, sizeof(mystring), pFile);
      continue;
    }
    // look for our combination of host and version
    std::string::size_type last_space_pos = line.rfind(' ');
    if (line.substr(0, space_pos) == srm_file_info.host && line.substr(last_space_pos+1) == srm_file_info.versionString()) {
      std::string::size_type space_pos2 = line.find(' ', space_pos+1);
      std::string::size_type space_pos3 = line.find(' ', space_pos2+1);
      if (space_pos3 != last_space_pos) {
        odlog(WARNING)<<"Bad format detected in file "<<srm_info_filename<<", in line "<<line<<std::endl;
        res = fgets (mystring, sizeof(mystring), pFile);
        continue;
      }
      int port_i;
      if (!stringtoint(line.substr(space_pos+1, space_pos2-(space_pos+1)), port_i)) {
        odlog(WARNING)<<"Can't convert string "<<line.substr(space_pos+1, space_pos2-(space_pos+1))<<" to int in file "<<srm_info_filename<<", line "<<line<<std::endl;
        res = fgets (mystring, sizeof(mystring), pFile);
        continue;
      }        
      srm_file_info.port = port_i;
      srm_file_info.protocol = line.substr(space_pos2+1, space_pos3-(space_pos2+1));
      fclose(pFile);
      filelock.unblock();
      return true;
    }
    res = fgets (mystring, sizeof(mystring), pFile);
  }
  fclose(pFile);
  filelock.unblock();
  return false;
}

void SRMInfo::putSRMFileInfo(const SRMFileInfo& srm_file_info) {
   
  struct stat fileStat;
  int err = stat( srm_info_filename.c_str(), &fileStat ); 
  if (0 != err || fileStat.st_size == 0) {
    if (0 != err && errno != ENOENT) {
      odlog(WARNING)<<"Error reading srm info file "<<srm_info_filename<<": "<<strerror(errno)<<std::endl;
      return;
    }
    
    // write new file
    filelock.block();
    
    FILE * pFile;
    pFile = fopen (srm_info_filename.c_str(), "w");
    if (pFile == NULL) {
      odlog(ERROR)<<"Error opening srm info file for writing "<<srm_info_filename<<": "<<strerror(errno)<<std::endl;
      filelock.unblock();
      return;
    }
    std::string header("# This file was automatically generated by ARC for caching SRM information.\n");
    header += "# Its format is lines with 4 entries separated by spaces:\n";
    header += "# hostname port protocol(gsi/gssapi) version\n#\n";
    header += "# This file can be freely edited, but it is not advisable while there\n";
    header += "# are on-going transfers. Comments begin with #\n#\n";
    header += srm_file_info.host + ' ' + inttostring(srm_file_info.port) + ' ' + srm_file_info.protocol + ' ' + srm_file_info.versionString() + '\n';                       
    fputs ((char*)header.c_str(), pFile);
    fclose (pFile);
    filelock.unblock();
    return;
  }

  // file already exists, so add this entry/replace existing entry
  filelock.block();
  FILE * pFile;
  char mystring [fileStat.st_size+1];
  pFile = fopen (srm_info_filename.c_str(), "r");
  if (pFile == NULL) {
    odlog(WARNING)<<"Error opening srm info file "<<srm_info_filename<<": "<<strerror(errno)<<std::endl;
    filelock.unblock();
    return;
  }

  // read in file
  std::vector<std::string> lines;
  
  char * res = fgets (mystring, sizeof(mystring), pFile);
  while (res) {
    std::string line(mystring);
    if (line.find('\n') != std::string::npos) line.resize(line.find('\n'));
    if (line.length() > 0 && line[0] == '#') {
      lines.push_back(line+'\n');
      res = fgets (mystring, sizeof(mystring), pFile);
      continue;
    }
    std::string::size_type space_pos = line.find(' ');
    if (space_pos == 0 || space_pos == std::string::npos) {
      if (line.length() > 0) 
        odlog(WARNING)<<"Bad format detected in file "<<srm_info_filename<<", in line "<<line<<std::endl;
      res = fgets (mystring, sizeof(mystring), pFile);
      continue;
    }
    // if any line contains our combination of host and version, ignore it
    std::string::size_type last_space_pos = line.rfind(' ');
    if (line.substr(0, space_pos) == srm_file_info.host && line.substr(last_space_pos+1) == srm_file_info.versionString()) {
      res = fgets (mystring, sizeof(mystring), pFile);
      continue;
    }
    lines.push_back(line+'\n');
    res = fgets (mystring, sizeof(mystring), pFile);
  }
  fclose(pFile);

  // write everything back to the file
  pFile = fopen (srm_info_filename.c_str(), "w");
  if (pFile == NULL) {
    odlog(WARNING)<<"Error opening srm info file for writing "<<srm_info_filename<<": "<<strerror(errno)<<std::endl;
    filelock.unblock();
    return;
  }
  
  for (std::vector<std::string>::iterator i = lines.begin(); i != lines.end(); i++) {
    fputs ((char*)i->c_str(), pFile);
  }
  std::string this_info = srm_file_info.host + ' ' + inttostring(srm_file_info.port) + ' ' + srm_file_info.protocol + ' ' + srm_file_info.versionString() + '\n';                       
  fputs ((char*)this_info.c_str(), pFile);
  fclose (pFile);
  filelock.unblock();
}
