#ifndef __DATAHANDLE_H__
#define __DATAHANDLE_H__

#include "../std.h"

#include <string>
#include <list>

#include "../misc/condition.h"
#include "../misc/globus_modules.h"
#include "datastatus.h"

#define MAX_PARALLEL_STREAMS 20
#define MAX_BLOCK_SIZE (1024*1024)

#ifndef _AIX
class DataPoint;
class DataBufferPar;
class DataCallback;
#else
#include "datapoint.h"
#include "databufferpar.h"
#include "datacallback.h"
#endif

/// DataHandle is kind of generalized file handle. Differently 
/// from file handle it does not support operations read() and 
/// write(). Instead it initiates operation and uses object of 
/// class DataBufferPar to pass actual data. It also provides 
/// other operations like querying parameters of remote object. 
/// It is used by higher-level classes DataMove and DataMovePar 
/// to provide data transfer service for application.
class DataHandle {
 private:
  DataHandle* instance;
  typedef DataHandle* (*constructor_t)(DataPoint* url);
  static std::list<constructor_t> protocols;
  static LockSimple protocols_lock;

 protected:
  static std::string utils_dir;
  virtual bool init_handle(void); 
  virtual bool deinit_handle(void); 
  DataHandle(void):instance(NULL) { };

 public:

  /// Constructor
  /// \param url_ URL. Should not be destroyed before DataHandle itself. 
  DataHandle(DataPoint* url_);
  /// Destructor. No comments.
  virtual ~DataHandle(void);

  /// Register new protocol. Any new instance of DataHandle will
  /// recognize it.
  static bool AddProtocol(constructor_t constructor);
  static void setUtilsDir(std::string dir);
  static DataHandle* CreateInstance(DataPoint* url_);
  DataHandle* Instance(void) { if(instance) return instance; return this; };
  const DataHandle* constInstance(void) const { if(instance) return instance; return this; };

  /// Start reading data from URL.
  /// Separate thread to transfer data will be created. No other operation
  /// can be performed while 'reading' is in progress.
  /// \param buffer operation will use this buffer to put information into. Should not be destroyed before stop_reading was called and returned.
  virtual DataStatus start_reading(DataBufferPar &buffer);
  /// Start writing data to URL.
  /// Separate thread to transfer data will be created. No other operation 
  /// can be performed while 'writing' is in progress.
  /// \param buffer operation will use this buffer to get information from. Should not be destroyed before stop_writing was called and returned.
  /// space_cb callback which is called if there is not enough to space storing data. Currently implemented only for file:/// URL.
  virtual DataStatus start_writing(DataBufferPar &buffer,DataCallback *space_cb = NULL);
  /// Stop reading. It MUST be called after corressponding start_reading 
  /// method. Either after whole data is transfered or to cancel transfer.
  /// Use 'buffer' object to find out when data is transfered.
  virtual DataStatus stop_reading(void);
  /// Same as stop_reading but for corresponding start_writing.
  virtual DataStatus stop_writing(void);
  /// Structure used in analyze() call. 
  /// \param bufsize returns suggested size of buffers to store data.
  /// \param bufnum returns suggested number of buffers.
  /// \param cache returns true if url is allowed to be cached.
  /// \param local return true if URL is accessed locally (file://)
  class analyze_t {
   public:
    long int bufsize;
    int bufnum;
    bool cache;
    bool local;
    bool readonly;
    analyze_t(void):bufsize(-1),bufnum(1),cache(true),local(false),readonly(true) { };
  };
  ///  Analyze url and provide hints.
  /// \param arg returns suggested values.
  virtual bool analyze(analyze_t &arg);
  /// Query remote server or local file system to check if object is
  /// accessible. If possible this function will also try to fill meta
  /// information about object in associated DataPoint.
  virtual DataStatus check(void);
  /// Remove/delete object at URL.
  virtual DataStatus remove(void);
  /// List files in directory or service (URL must point to directory/group/service access point).
  /// \param files will contain list of file names and optionally their attributes.
  /// \param long_list if true find details for each file
  /// \param resolve if true resolve location URLs (for meta-URLs only).
  /// \param metadata if true fill all possible metadata in files
  virtual DataStatus list_files(std::list<DataPoint::FileInfo> &files,bool long_list = false,bool resolve = false,bool metadata = false);
  /// Returns true if URL can accept scatterd data (like arbitrary access to local file) for 'writing' operation.
  virtual bool out_of_order(void);
  /// Allow/disallow DataHandle to produce scattered data during 'reading' operation.
  /// \param v true if allowed.
  virtual void out_of_order(bool v);
  /// Allow/disallow to make check for existance of remote file (and probably other checks too) before initiating 'reading' and 'writing' operations.
  /// \param v true if allowed (default is true).
  virtual void additional_checks(bool v);
  /// Check if additional checks before 'reading' and 'writing' will be performed.
  virtual bool additional_checks(void);
  /// Allow/disallow heavy security during data transfer.
  /// \param v true if allowed (default is true only for gsiftp://).
  virtual void secure(bool v);
  /// Check if heavy security during data transfer is allowed.
  virtual bool secure(void);
  /// Request passive transfers for FTP-like protocols.
  /// \param true to request.
  virtual void passive(bool v);
  /// Returns reason of transfer failure, as reported by callbacks.
  /// This could be different from the failure returned by the methods themselves.
  virtual DataStatus failure_reason(void);
  /// Set range of bytes to retrieve. Default values correspond to whole file. 
  virtual void range(unsigned long long int start = 0,unsigned long long end = 0);
};

class DataHandleCommon: public DataHandle {
 protected:
  DataPoint* url; /* full url with locations, etc. */
  DataBufferPar* buffer; /* pointer to buffer for data */
  std::string c_url;   /* current url with options stripped */ 
  bool cacheable;
  bool linkable;
  bool is_secure;
  bool force_secure;
  bool force_passive;
  bool reading;
  bool writing;
  bool no_checks;
  /* true if allowed to produce scattered data (non-sequetial offsets) */
  bool allow_out_of_order;
  int transfer_streams;
  unsigned long long int range_start;
  unsigned long long int range_end;

  DataStatus failure_code; /* filled by callback methods */

  virtual bool init_handle(void); 
  virtual bool deinit_handle(void); 
 public:
  static DataHandle* CreateInstance(DataPoint* url_);
  DataHandleCommon(DataPoint* url_);
  virtual ~DataHandleCommon(void);
  virtual bool out_of_order(void);
  virtual void out_of_order(bool v);
  virtual void additional_checks(bool v);
  virtual bool additional_checks(void);
  virtual void secure(bool v);
  virtual bool secure(void);
  virtual void passive(bool v);
  virtual DataStatus failure_reason(void);
  virtual void range(unsigned long long int start = 0,unsigned long long end = 0);
  virtual DataStatus start_reading(DataBufferPar &buffer);
  virtual DataStatus start_writing(DataBufferPar &buffer,DataCallback *space_cb = NULL);
  virtual DataStatus stop_reading(void);
  virtual DataStatus stop_writing(void);
  virtual bool analyze(analyze_t &arg);
  virtual DataStatus check(void);
  virtual DataStatus remove(void);
  virtual DataStatus list_files(std::list<DataPoint::FileInfo> &files,bool long_list = false,bool resolve = false,bool metadata = false);
};

#endif
